@@ -6,12 +6,14 @@
  */
 package org.gridsuite.gridpy;
 
+import ch.qos.logback.classic.Level;
+import ch.qos.logback.classic.Logger;
 import com.powsybl.ieeecdf.converter.IeeeCdfNetworkFactory;
 import com.powsybl.iidm.export.Exporters;
 import com.powsybl.iidm.import_.Importers;
-import com.powsybl.iidm.network.Bus;
-import com.powsybl.iidm.network.Network;
+import com.powsybl.iidm.network.*;
 import com.powsybl.iidm.network.test.EurostagTutorialExample1Factory;
+import com.powsybl.iidm.network.util.ConnectedComponents;
 import com.powsybl.loadflow.LoadFlow;
 import com.powsybl.loadflow.LoadFlowParameters;
 import com.powsybl.loadflow.LoadFlowResult;
@@ -33,16 +35,12 @@ import org.graalvm.nativeimage.c.type.CCharPointerPointer;
 import org.graalvm.nativeimage.c.type.CDoublePointer;
 import org.graalvm.nativeimage.c.type.CTypeConversion;
 import org.graalvm.word.PointerBase;
+import org.graalvm.word.WordFactory;
 import org.slf4j.LoggerFactory;
-import ch.qos.logback.classic.Level;
-import ch.qos.logback.classic.Logger;
 
 import java.nio.file.Paths;
-import java.util.Collection;
-import java.util.List;
-import java.util.Objects;
+import java.util.*;
 import java.util.stream.Collectors;
-import java.util.stream.Stream;
 
 import static org.gridsuite.gridpy.GridPyApiHeader.*;
 
@@ -156,18 +154,22 @@ public final class GridPyApi {
         freeArrayPointer(componentResultArrayPtr);
     }
 
+    private static void fillBus(Bus bus, BusPointer busPtr) {
+        busPtr.setId(CTypeConversion.toCString(bus.getId()).get());
+        busPtr.setVoltageMagnitude(bus.getV());
+        busPtr.setVoltageAngle(bus.getAngle());
+        busPtr.setComponentNum(ConnectedComponents.getCcNum(bus));
+    }
+
     @CEntryPoint(name = "getBusArray")
-    public static ArrayPointer<BusPointer> getBusArray(IsolateThread thread, ObjectHandle networkHandle, boolean busBreakerView) {
+    public static ArrayPointer<BusPointer> getBusArray(IsolateThread thread, ObjectHandle networkHandle) {
         Network network = ObjectHandles.getGlobal().get(networkHandle);
-        Stream<Bus> busStream = busBreakerView ? network.getBusBreakerView().getBusStream() : network.getBusView().getBusStream();
-        List<Bus> buses = busStream.collect(Collectors.toList());
+        List<Bus> buses = network.getBusView().getBusStream().collect(Collectors.toList());
         BusPointer busesPtr = UnmanagedMemory.calloc(buses.size() * SizeOf.get(BusPointer.class));
         for (int index = 0; index < buses.size(); index++) {
             Bus bus = buses.get(index);
             BusPointer busPtr = busesPtr.addressOf(index);
-            busPtr.setId(CTypeConversion.toCString(bus.getId()).get());
-            busPtr.setVoltageMagnitude(bus.getV());
-            busPtr.setVoltageAngle(bus.getAngle());
+            fillBus(bus, busPtr);
         }
         return allocArrayPointer(busesPtr, buses.size());
     }
@@ -178,6 +180,78 @@ public final class GridPyApi {
         freeArrayPointer(busArrayPointer);
     }
 
+    @CEntryPoint(name = "getGeneratorArray")
+    public static ArrayPointer<GeneratorPointer> getGeneratorArray(IsolateThread thread, ObjectHandle networkHandle) {
+        Network network = ObjectHandles.getGlobal().get(networkHandle);
+        List<Generator> generators = network.getGeneratorStream().collect(Collectors.toList());
+        GeneratorPointer generatorPtr = UnmanagedMemory.calloc(generators.size() * SizeOf.get(GeneratorPointer.class));
+        for (int index = 0; index < generators.size(); index++) {
+            Generator generator = generators.get(index);
+            GeneratorPointer generatorPtrI = generatorPtr.addressOf(index);
+            generatorPtrI.setId(CTypeConversion.toCString(generator.getId()).get());
+            generatorPtrI.setTargetP(generator.getTargetP());
+            generatorPtrI.setMaxP(generator.getMaxP());
+            generatorPtrI.setMinP(generator.getMinP());
+            VoltageLevel vl = generator.getTerminal().getVoltageLevel();
+            generatorPtrI.setNominalVoltage(vl.getNominalV());
+            generatorPtrI.setCountry(CTypeConversion.toCString(vl.getSubstation().getCountry().map(Country::name).orElse(null)).get());
+            Bus bus = generator.getTerminal().getBusView().getBus();
+            if (bus != null) {
+                BusPointer busPtr = UnmanagedMemory.calloc(SizeOf.get(BusPointer.class));
+                fillBus(bus, busPtr);
+                generatorPtrI.setBus(busPtr);
+            }
+        }
+        return allocArrayPointer(generatorPtr, generators.size());
+    }
+
+    @CEntryPoint(name = "freeGeneratorArray")
+    public static void freeGeneratorArray(IsolateThread thread, ArrayPointer<GeneratorPointer> generatorArrayPtr) {
+        for (int index = 0; index < generatorArrayPtr.getLength(); index++) {
+            GeneratorPointer generatorPtrI = generatorArrayPtr.getPtr().addressOf(index);
+            // don't need to free char* from id field as it is done by python
+            if (generatorPtrI.getBus().isNonNull()) {
+                UnmanagedMemory.free(generatorPtrI.getBus());
+            }
+        }
+        freeArrayPointer(generatorArrayPtr);
+    }
+
+    @CEntryPoint(name = "getLoadArray")
+    public static ArrayPointer<LoadPointer> getLoadArray(IsolateThread thread, ObjectHandle networkHandle) {
+        Network network = ObjectHandles.getGlobal().get(networkHandle);
+        List<Load> loads = network.getLoadStream().collect(Collectors.toList());
+        LoadPointer loadPtr = UnmanagedMemory.calloc(loads.size() * SizeOf.get(LoadPointer.class));
+        for (int index = 0; index < loads.size(); index++) {
+            Load load = loads.get(index);
+            LoadPointer loadPtrI = loadPtr.addressOf(index);
+            loadPtrI.setId(CTypeConversion.toCString(load.getId()).get());
+            loadPtrI.setP0(load.getP0());
+            VoltageLevel vl = load.getTerminal().getVoltageLevel();
+            loadPtrI.setNominalVoltage(vl.getNominalV());
+            loadPtrI.setCountry(CTypeConversion.toCString(vl.getSubstation().getCountry().map(Country::name).orElse(null)).get());
+            Bus bus = load.getTerminal().getBusView().getBus();
+            if (bus != null) {
+                BusPointer busPtr = UnmanagedMemory.calloc(SizeOf.get(BusPointer.class));
+                fillBus(bus, busPtr);
+                loadPtrI.setBus(busPtr);
+            }
+        }
+        return allocArrayPointer(loadPtr, loads.size());
+    }
+
+    @CEntryPoint(name = "freeLoadArray")
+    public static void freeLoadArray(IsolateThread thread, ArrayPointer<LoadPointer> loadArrayPtr) {
+        for (int index = 0; index < loadArrayPtr.getLength(); index++) {
+            LoadPointer loadPtrI = loadArrayPtr.getPtr().addressOf(index);
+            // don't need to free char* from id field as it is done by python
+            if (loadPtrI.getBus().isNonNull()) {
+                UnmanagedMemory.free(loadPtrI.getBus());
+            }
+        }
+        freeArrayPointer(loadArrayPtr);
+    }
+
     @CEntryPoint(name = "updateSwitchPosition")
     public static boolean updateSwitchPosition(IsolateThread thread, ObjectHandle networkHandle, CCharPointer id, boolean open) {
         Network network = ObjectHandles.getGlobal().get(networkHandle);
@@ -194,9 +268,12 @@ public final class GridPyApi {
 
     @CEntryPoint(name = "getNetworkElementsIds")
     public static ArrayPointer<CCharPointerPointer> getNetworkElementsIds(IsolateThread thread, ObjectHandle networkHandle, ElementType elementType,
-                                                                                          double nominalVoltage, boolean mainCc) {
+                                                                          CDoublePointer nominalVoltagePtr, int nominalVoltageCount,
+                                                                          CCharPointerPointer countryPtr, int countryCount, boolean mainCc) {
         Network network = ObjectHandles.getGlobal().get(networkHandle);
-        List<String> elementsIds = NetworkUtil.getElementsIds(network, elementType, nominalVoltage, mainCc);
+        Set<Double> nominalVoltages = new HashSet<>(CTypeUtil.createDoubleList(nominalVoltagePtr, nominalVoltageCount));
+        Set<String> countries = new HashSet<>(CTypeUtil.createStringList(countryPtr, countryCount));
+        List<String> elementsIds = NetworkUtil.getElementsIds(network, elementType, nominalVoltages, countries, mainCc);
         CCharPointerPointer elementsIdsPtr = UnmanagedMemory.calloc(elementsIds.size() * SizeOf.get(CCharPointerPointer.class));
         for (int i = 0; i < elementsIds.size(); i++) {
             elementsIdsPtr.addressOf(i).write(CTypeConversion.toCString(elementsIds.get(i)).get());
@@ -321,21 +398,40 @@ public final class GridPyApi {
         SensitivityAnalysisResultContext resultContext = ObjectHandles.getGlobal().get(sensitivityAnalysisResultContextHandle);
         String contingencyId = CTypeConversion.toJavaString(contingencyIdPtr);
         Collection<SensitivityValue> sensitivityValues = resultContext.getSensitivityValues(contingencyId);
-        MatrixPointer matrixPtr = UnmanagedMemory.calloc(SizeOf.get(MatrixPointer.class));
         if (sensitivityValues != null) {
             CDoublePointer valuePtr = UnmanagedMemory.calloc(resultContext.getRowCount() * resultContext.getColumnCount() * SizeOf.get(CDoublePointer.class));
             for (SensitivityValue sensitivityValue : sensitivityValues) {
                 IndexedSensitivityFactor indexedFactor = (IndexedSensitivityFactor) sensitivityValue.getFactor();
                 valuePtr.addressOf(indexedFactor.getRow() * resultContext.getColumnCount() + indexedFactor.getColumn()).write(sensitivityValue.getValue());
             }
+            MatrixPointer matrixPtr = UnmanagedMemory.calloc(SizeOf.get(MatrixPointer.class));
             matrixPtr.setRowCount(resultContext.getRowCount());
             matrixPtr.setColumnCount(resultContext.getColumnCount());
             matrixPtr.setValues(valuePtr);
-        } else {
-            matrixPtr.setRowCount(0);
-            matrixPtr.setColumnCount(0);
+            return matrixPtr;
         }
-        return matrixPtr;
+        return WordFactory.nullPointer();
+    }
+
+    @CEntryPoint(name = "getReferenceFlows")
+    public static MatrixPointer getReferenceFlows(IsolateThread thread, ObjectHandle sensitivityAnalysisResultContextHandle,
+                                                  CCharPointer contingencyIdPtr) {
+        SensitivityAnalysisResultContext resultContext = ObjectHandles.getGlobal().get(sensitivityAnalysisResultContextHandle);
+        String contingencyId = CTypeConversion.toJavaString(contingencyIdPtr);
+        Collection<SensitivityValue> sensitivityValues = resultContext.getSensitivityValues(contingencyId);
+        if (sensitivityValues != null) {
+            CDoublePointer valuePtr = UnmanagedMemory.calloc(resultContext.getColumnCount() * SizeOf.get(CDoublePointer.class));
+            for (SensitivityValue sensitivityValue : sensitivityValues) {
+                IndexedSensitivityFactor indexedFactor = (IndexedSensitivityFactor) sensitivityValue.getFactor();
+                valuePtr.addressOf(indexedFactor.getColumn()).write(sensitivityValue.getFunctionReference());
+            }
+            MatrixPointer matrixPtr = UnmanagedMemory.calloc(SizeOf.get(MatrixPointer.class));
+            matrixPtr.setRowCount(1);
+            matrixPtr.setColumnCount(resultContext.getColumnCount());
+            matrixPtr.setValues(valuePtr);
+            return matrixPtr;
+        }
+        return WordFactory.nullPointer();
     }
 
     @CEntryPoint(name = "destroyObjectHandle")
